import copy
import httplib
import json
import time
from datetime import datetime
import requests
from common import *
import logging
import traceback
import flask
from flask import jsonify

application = flask.Flask(__name__)

# Create logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
# Handler
LOG_FILE = '/opt/python/log/app.log'
if not os.path.exists(os.path.dirname(LOG_FILE)):
    os.makedirs(os.path.dirname(LOG_FILE))
handler = logging.FileHandler(LOG_FILE)
handler.setLevel(logging.INFO)
# Formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# Add Formatter to Handler
handler.setFormatter(formatter)
# add Handler to Logger
logger.addHandler(handler)

from apscheduler.schedulers.background import BackgroundScheduler

scheduler = BackgroundScheduler()


def calculate_sha1(file_full_path):
    import hashlib
    BUF_SIZE = 65536
    try:
        sha1 = hashlib.sha1()
        with open(file_full_path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)
        return sha1.hexdigest()
    except Exception:
        logger.error("Error in calculate_sha1 file_path : %s " % file_full_path)
        return None


def get_api_key_header():
    headers = {'X-Api-Key': VV_API_KEY}
    return headers


def vv_file_query_api(sha1):
    json_response = None
    request_url = VALKYRIE_FILE_QUERY_API + sha1
    response = requests.request('GET', request_url, headers=get_api_key_header())
    if response.status_code == httplib.OK:
        json_response = json.loads(response.text)
    return json_response


def vv_file_upload_api(file_path):
    json_response = None
    if os.path.getsize(file_path) >> 20 < 25:
        with open(file_path, 'rb') as f:
            file_data = {
                "file_data": f
            }
            response = requests.post(VALKYRIE_FILE_UPLOAD_API, files=file_data, timeout=(30, 30), headers=get_api_key_header())
            if response.status_code == httplib.OK:
                json_response = json.loads(response.text)
    return json_response


def wait_for_analysis(sha1):
    verdict = None
    file_status = "UPLOAD_TIMEOUT"
    time_passed = 0

    while verdict is None and time_passed <= UPLOAD_WAIT_TIMEOUT:
        query_response = vv_file_query_api(sha1)
        if query_response:
            query_rcode = query_response.get("return_code")
            if query_rcode == VVReturnCodes.SUCCESS:
                # File exists on VV
                verdict = query_response.get("verdict")
                file_status = "UPLOAD_SUCCESS"
                break
            elif query_rcode in VVReturnCodes.REQUEST_LIMIT_REACHED:
                # Request limit reached abort process
                file_status = "QUERY_LIMIT_REACHED"
                break
        # Sleep before checking again
        time.sleep(UPLOAD_CHECK_INTERVAL)
        time_passed += UPLOAD_CHECK_INTERVAL

    return verdict, file_status


def set_vv_tags(s3_object, sha1, verdict, file_status):
    curr_tags = s3_client.get_object_tagging(Bucket=s3_object.bucket_name, Key=s3_object.key)["TagSet"]
    new_tags = copy.copy(curr_tags)
    for tag in curr_tags:
        if tag["Key"] in [VV_STATUS_TAG, VV_TIMESTAMP_TAG, VV_SHA1_TAG, VV_VERDICT_TAG]:
            new_tags.remove(tag)

    new_tags.append({"Key": VV_SHA1_TAG, "Value": str(sha1)})
    new_tags.append({"Key": VV_VERDICT_TAG, "Value": str(verdict)})
    new_tags.append({"Key": VV_STATUS_TAG, "Value": str(file_status)})
    new_tags.append({"Key": VV_TIMESTAMP_TAG, "Value": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")})
    s3_client.put_object_tagging(
        Bucket=s3_object.bucket_name,
        Key=s3_object.key,
        Tagging={"TagSet": new_tags}
    )


def start_scan_process(sha1, file_path):
    query_response = vv_file_query_api(sha1)
    verdict = None
    file_status = ""

    if query_response:
        query_rcode = query_response.get("return_code")
        if query_rcode == VVReturnCodes.SUCCESS:
            # File exists on VV
            verdict = query_response.get("verdict")
            file_status = "QUERY_SUCCESS"
        elif query_rcode == VVReturnCodes.FILE_NOT_EXISTS:
            # File does not exists on VV so upload file for scan
            upload_response = vv_file_upload_api(file_path)
            if upload_response:
                upload_rcode = upload_response.get("return_code")
                if upload_rcode == VVReturnCodes.SUCCESS:
                    # Upload success wait for completion
                    verdict, file_status = wait_for_analysis(sha1)
                elif upload_rcode in VVReturnCodes.REQUEST_LIMIT_REACHED:
                    # Defined request limit is reached for given api key, so skip file
                    file_status = "UPLOAD_LIMIT_REACHED"
                else:
                    # Unexpected response returned, skip file
                    file_status = "UPLOAD_FAIL_%s" % upload_rcode
            else:
                pass
        elif query_rcode in VVReturnCodes.REQUEST_LIMIT_REACHED:
            # Defined request limit is reached for given api key, so skip file
            file_status = "QUERY_LIMIT_REACHED"
        else:
            # Unexpected response returned, skip file
            file_status = "QUERY_FAIL_%s" % query_rcode

    return verdict, file_status


def download_s3_object(s3_object, local_prefix):
    local_path = "%s/%s/%s" % (local_prefix, s3_object.bucket_name, s3_object.key)
    create_dir(os.path.dirname(local_path))
    s3_object.download_file(local_path)
    return local_path

def change_acl(bucket, key, acl='private'):
    object_acl = s3.ObjectAcl(bucket, key)
    response = object_acl.put(ACL=acl)
    return response


def get_s3_key_list(name):
    s3_key_list = []
    b = s3_client.list_objects_v2(Bucket=name, FetchOwner=True, MaxKeys=1000)
    # capture metadata about objects in bucket
    while True:
        if b.has_key('Contents'):
            for z in b['Contents']:
                size = str(z['Size'])
                skey = str(z['Key'])
                if size and int(size) > 0 and int(size) >> 20 < 25:  # 0 size means empty file or folder name
                    s3_key_list.append(skey)
        is_truncated = b["IsTruncated"]
        if not is_truncated:
            break
        else:
            b = s3_client.list_objects_v2(Bucket=name, FetchOwner=True, MaxKeys=1000, ContinuationToken=b["NextContinuationToken"])

    return s3_key_list


def process_list(s3_key_list):
    for key in s3_key_list:
        s3_object = s3.Object(VV_BUCKET_NAME, key)
        curr_tags = s3_client.get_object_tagging(Bucket=s3_object.bucket_name, Key=s3_object.key)["TagSet"]
        process_file_flag = True
        sha1 = ''
        for tag in curr_tags:
            if tag["Key"] == VV_VERDICT_TAG:
                if tag["Value"] in [VERDICT_MALWARE, VERDICT_CLEAN, VERDICT_PUA]:
                    process_file_flag = False
            elif tag["Key"] == VV_SHA1_TAG:
                sha1 = tag["Value"]
        if process_file_flag:
            process_file(s3_object, sha1)


def process_file(s3_object, sha1):
    file_path = download_s3_object(s3_object, "/tmp")
    if len(sha1) == 0:
        sha1 = calculate_sha1(file_path)

    if VV_CHANGE_PERMISSION:  #change file permissions to private
        try:
            logger.info("Change permission is true , changing permission to private")
            response = change_acl(s3_object.bucket_name, s3_object.key, 'private')
            if response['ResponseMetadata']['HTTPStatusCode'] == httplib.OK:
                logger.info("Changed permission to private successfully")
        except:
            logger.error("Exception in setting permissions to private err %s" % traceback.format_exc())

    verdict, file_status = start_scan_process(sha1, file_path)
    verdict = VALKYRIE_VERDICT_MAP.get(verdict, "UNKNOWN")
    logger.info("Scan of s3://%s resulted - SHA1: %s VERDICT: %s STATUS: %s\n" % (os.path.join(s3_object.bucket_name, s3_object.key), sha1, verdict, file_status))
    set_vv_tags(s3_object, sha1, verdict, file_status)

    if VV_CHANGE_PERMISSION:  #revert file permissions if in allowed verdicts
        try:
            if VV_ALLOWED_VERDICTS and len(VV_ALLOWED_VERDICTS) > 0:
                allowed_verdict_text = VV_ALLOWED_VERDICTS.upper()
            else:
                allowed_verdict_text = "CLEAN,UNDETECTED"
            if verdict in allowed_verdict_text:
                logger.info("Change permission is true , changing permission to public-read")
                response = change_acl(s3_object.bucket_name, s3_object.key, 'public-read')
                if response['ResponseMetadata']['HTTPStatusCode'] == httplib.OK:
                    logger.info("Changed permission to public-read successfully")
            else:
                logger.info("File verdict is not in allowed verdicts file verdict:%s allowed verdicts:%s" % (verdict, VV_ALLOWED_VERDICTS))
        except:
            logger.error("Exception in reverting permissions err %s" % traceback.format_exc())
    try:
        os.remove(file_path)
    except OSError:
        pass


def check_files():
    logger.info("VV-Check starting at %s\n" % (datetime.utcnow().strftime("%Y/%m/%d %H:%M:%S UTC")))
    try:
        s3_key_list = get_s3_key_list(VV_BUCKET_NAME)
        logger.info("Total files %s " % len(s3_key_list))
        process_list(s3_key_list)
        logger.info("VV-Check finished at %s\n" % datetime.utcnow().strftime("%Y/%m/%d %H:%M:%S UTC"))
    except:
        logger.error("VV-Check Failed: \n %s " + traceback.format_exc())


@application.route('/scheduled', methods=["POST"])
def scheduled():
    logger.info("Starting check_files")
    check_files()
    logger.info("Finished check_files")
    return jsonify({"message": "Finished scheduled task"})


job = scheduler.add_job(check_files, 'interval', minutes=int(VV_CRONJOB_INTERVAL))
scheduler.start()

logger.info("Started application with following settings..")
logger.info("VV_CRONJOB_INTERVAL:%s" % VV_CRONJOB_INTERVAL)
logger.info("VV_BUCKET_NAME:%s" % VV_BUCKET_NAME)
logger.info("VV_API_KEY:%s" % VV_API_KEY)

if __name__ == '__main__':
    #check_files() # TODO remove
    application.run(host='0.0.0.0')
